#
# pgp.tcl
#	PGP 2.6/PGP 5.0/GPG support for exmh.
#	Orginally contributed by Allan Downey
#	Updated by Stefan Monnier, Anders Klemets, William Sproule,
#	Chris Garrigues and Ben Escoto
#

# future:
# - rewrite PGP decrypt, to deal with keys, mime stuff, etc...
# - encrypt pgp parts, similar to the automatic quote-printable
# - split big functions
# - drop misc_displaytext
# - add a "init PGP" command
# - add some key handling commands
# - keep track of who has your private key (for revocation purposes)

# Pgp_Init is in extrasInit.tcl
# to avoid auto-loading this whole file.

# $Log: pgp.tcl,v $
# Revision 1.16  1998/01/22  00:47:01  bwelch
#     Fixed Pgp_Setup to use /dev/tty instead of /dev/console
#
# Revision 1.15  1997/12/22  20:53:46  bwelch
# File_Delete
#
# Revision 1.14  1997/07/25  17:14:09  bwelch
# Trapped error messages from PGP xterms.
#
# Revision 1.13  1997/07/12  23:06:28  bwelch
#     Added expecttk support for PGP handling.
#     Fixed recursive failures with some PGP messages.
#
# Revision 1.12  1997/06/03  18:37:00  bwelch
# Added doinc flag to Inc_Presort
# Major cleanup
#
# Revision 1.11  1997/01/25  06:21:16  bwelch
# Added Quote module support
#
# Revision 1.10  1997/01/25  05:34:08  bwelch
#     Added Pgp_GetTextAttributes
# Changed display of PGP messages
#
# Revision 1.9  1996/12/21  00:58:08  bwelch
# Fixed PGP menu to support both email and www key fetching.
# Guard against missing signature parts in multipart/signed
#
# Revision 1.8  1996/12/02  21:11:11  bwelch
# Moved New Key button to be inside the help documents.
#
# Revision 1.7  1996/12/01  20:16:47  bwelch
# Added mutipart/security support.
# (Ben Escoto and Chris Garrigues)
#
# Revision 1.6  1996/03/22  18:44:34  bwelch
# Changed graphic part separator size in PGP from 5 to 6 to avoid
#     downloading a new font for this case.
#
# Revision 1.5  1995/09/28  04:11:10  bwelch
# Fixed "hasfcc" check in PGP.
#
# Revision 1.4  1995/06/30  18:32:40  bwelch
# Upcase PGP mail headers
#
# Revision 1.3  1995/06/09  20:57:06  bwelch
# Added ChoosePrivateKey
#
# Revision 1.2  1995/05/24  06:01:38  bwelch
# Added Pgp_SetMyName to choose private key name
#
# Revision 1.1  1995/05/24  01:48:03  bwelch
# Initial revision
#
# Revision 1.21  1995/04/15  18:17:01  welch
# Introduced msg(path)
#
# Revision 1.20  1995/03/22  22:17:30  welch
# Changed exmh.PGP.help to help.PGP
#
# Revision 1.19  1995/03/22  18:53:52  welch
# More new code from Stefan
#
# Revision 1.1  1994/12/17  20:18:49  monnier
# Initial revision
#


# Pgp_Setup is invoked by the "Make Key" button on the PGP Setup Help page
# It searches for versions on the system, which are supported exmh
# (at the moment PGP 2.6.x, PGP 5.0 (new) and GPG (new)).
# Pgp_Setup then configures the version found on your system.
#
# If there are more than one version (2.6 and 5.0 and GPG) installed,
# then a dialog box pops up and asks the user for the version, he/she wants
# to use:
# He/She can choose a single version, two of them or all.
#
# If a user likes to communicate with both PGP 2.6 and PGP 5.x users,
# there should be both versions installed and configured on the system.
# If a user wants to communicate with GPG users,
# there must be GPG installed on the system, because PGP2.6/PGP5.0 -- GPG
# interoperation is still problematic
#
#
# added pgp5 support
# added gpg support
proc Pgp_Setup {} {
    global pgp pgp5 gpg env
    
    PgpSetPath

    #
    # Executables
    set pgp(exe)  pgp
    set pgp5(exe) pgpk
    set gpg(exe)  gpg

    foreach program [list pgp pgp5 gpg] {
    	foreach path [split $env(PATH) ":"] {
	    if [file executable "$path/[set ${program}(exe)]"] {
		lappend installed $program
		break
	    }
	}
    }
    
    if {![info exists installed]} {
	Misc_DisplayText "PGP Setup" \
"A PGP executable is not in your PATH.
You'll have to find it (or install it) before
I can do anything for you."
        return
    }

    #
    # dialog box, if there are more than one versions installed
    #
    set buttontext(pgp) "PGP 2.6"
    set buttontext(pgp5) "PGP 5.0"
    set buttontext(gpg) "GPG"

    if { [llength $installed] >= 2 } {
	set t [toplevel .hugo]
	wm title $t "Which version?"
	wm resizable $t 0 0
	set f [frame $t.frame1]
	pack $f -side top
	set l [label $f.pic -bitmap questhead]
	pack $l -side left -padx 5
	set m [message $f.msg -justify left -text \
"There are three versions installed on your system. Which version(s) do you want to use to generate your key?"]
	pack $m -side left
	set f [frame $t.frame2]
	pack $f -side top
	global value
	foreach v $installed {
	    set value($v) 0
	    set c [checkbutton $f.$v -variable value($v) \
					-text [set buttontext($v)] ]
	    pack $c -side left
	}
	set b [button $f.ok -text OK -command "destroy $t"]
	pack $b -side left
	
	tkwait window $t
	
	foreach v $installed {
	    if {[set value($v)]} {
		lappend versions $v
	    }
	}
	if {![info exists versions]} {
	    return
	}
    } else {
	switch -regexp -- $pgpexec {
		pgpk { set versions pgp5 }
		pgp { set versions pgp }
		gpg { set versions gpg }
	}
    }

    # setup the directory
    if {![info exists env(PGPPATH)]} {
	set env(PGPPATH) "$env(HOME)/.pgp"
    }
    catch { exec mkdir $env(PGPPATH) }

    # GPG
    catch { exec mkdir $env(HOME)/.gnupg } 
    
    # setup config files
    foreach version $versions {
	switch $version {
	     pgp5 { exec touch "$env(PGPPATH)/pgp.cfg" }
	     pgp  { exec touch "$env(PGPPATH)/config.txt" }
	     gpg  { exec touch "$env(HOME)/.gnupg/options" }
	}
    }

    # make the key pair(s) and self sign it/them
    foreach version $versions {
	switch $version {
	     gpg {
		if [catch {
		    exec xterm -title "GPG Setup" -e sh -c {
			cd ${HOME}/.gnupg
			gpg --gen-key
 			} >& /dev/tty
		    } error] {
		    Misc_DisplayText "GPG Setup Error" \
"An error occurred while trying to generate your key:
$error
please try these commands at your unix shell
to generate your key and self sign it:
	gpg --gen-key
Then restart exmh to enable its GPG support.
"
        	    return
		}
	     }
	     pgp5 {
		if [catch {
		    exec xterm -title "PGP5 Setup" -e sh -c {
			cd ${PGPPATH}
			rm -f pubring.bak
			pgpk -g
			} >& /dev/tty
		    } error] {
		    Misc_DisplayText "PGP5 Setup Error" \
"An error occurred while trying to generate your key:
$error
please try these commands at your unix shell
to generate your key and self sign it:
	pgpk -g
Then restart exmh to enable its PGP5 support.
"
        	    return
		}
	     }
	     pgp {
    		if [catch {
		    exec xterm -title "PGP Setup" -e sh -c {
	    	    	cd ${PGPPATH}
	    	    	rm -f pubring.bak
	    	    	pgp -kg
			# automatic keysigning
	    	    	#rm -f pubring.bak
	    	    	#pgp +verbose=0 +force=on -ks "" -u "" \
 		    	} >& /dev/tty
		    } error] {
		    Misc_DisplayText "PGP Setup Error" \
"An error occurred while trying to generate your key:
$error
please try these commands at your unix shell
to generate your key and self sign it:
	pgp -kg
	pgp +verbose=0 +force=on -ks \"\" -u \"\"
Then restart exmh to enable its PGP support.
"
        	    return
    	    	}
	    }
	}
    }

    # init the pgp support if necessary
    foreach version $versions {
    	if {![set ${version}(enabled)]} {
		Pgp_Init
    	} else {
	    switch $version {
		pgp5 { set ${version}(secring) $pgp(pgppath)/secring.skr }
		pgp  { set ${version}(secring) $pgp(pgppath)/secring.pgp }
		gpg  { set ${version}(secring) $gpg(gpgpath)/secring.gpg }
	    }
	    set ${version}(privatekeys)\
		[PgpExec_KeyList $version "" [set ${version}(secring)]]
    	}
    }
    
    if {![file exists "$env(PGPPATH)/pubring.bak"]} {
        return
    }

    # send the key to the keyservers; 5.0 prompts for it and for gpg there is
    # no keyserver yet ?!
    #
    foreach version $versions {
	if { $version == "pgp" } {
	    set tmpfile [Mime_TempFile "pgp"]
	    PgpExec_GetKeys pgp [lindex [lindex $pgp(privatekeys) 0] 0] $tmpfile
    	    Misc_Send $pgp(keyserver) ADD $tmpfile \
		"content-type: application/pgp; format=keys-only"
    	    File_Delete $tmpfile
    	}
    }
}

#
proc Pgp_Help {  } {
    Help PGP	;# Make Key button is embedded in the HTML
}
proc Pgp_HelpOld {} {
    global exmh
    set label "Help about setting up PGP"
    if [Exwin_Toplevel .pgphelp "PGP Help" PgpHelp] {
	Widget_Label .pgphelp.but label {left fill} -text $label
	Widget_AddBut .pgphelp.but setup "Make Key" [list Pgp_Setup]
    
	set t [Widget_Text .pgphelp 30 -setgrid true]
	Ftoc_ColorConfigure $t
	$t insert insert "EXMH Version: $exmh(version)\n\n"
	if [catch {open "$exmh(library)/help.PGP" r} in] {
	    $t insert insert "Cannot find file exmh.PGP.help to display"
	    $t configure -state disabled
	} else {
	    $t insert insert [read $in]
	}
    }
}

# does not support pgp5/gpg!
# display the message without keeping the decrypted form
proc Pgp_View { {file {}} } {
    global exmh msg exwin pgp

    if {$file == {}} { set file $msg(path) }

    if $pgp(keeppass) {
	PgpExec pgp [list -m $file] message $pgp(myname)
	Misc_DisplayText "PGP view $exmh(folder)/$msg(id)" $message 40
    } else {
	PgpExec pgp [list -m $file] message $pgp(myname)
    }

    return 1
}

# does not support pgp5/gpg!
# decrypt the current message
proc Pgp_Decrypt {  } {
    global exmh msg mhProfile exwin pgp

    set file $msg(path)
    set pgpfile [Mime_TempFile "decrypt"]

    Exmh_Status "pgp $exmh(folder)/$msg(id)"
    PgpExec pgp [list $file -o $pgpfile] message $pgp(myname)
    if {$message != {}} {
	Misc_DisplayText "PGP decrypt $exmh(folder)/$msg(id)" $message
    }
    set t $exwin(mtext)

    set orig [open $file r]
    if [catch {
	set mess [open $pgpfile r]
    } err] {
	return 0
    }
    set outfile [Mime_TempFile "decypted"]
    set comb [open $outfile w 0600]

    if {![file exists "$env(PGPPATH)/pubring.bak"]} {
	return
    }

		gpg  { set ${version}(secring) $gpg(gpgpath)/secring.gpg }
    while {[gets $orig line] != -1} {
	if {[regexp {^-+BEGIN PGP MESSAGE} $line]} {
	    puts -nonewline $comb [read $mess]
	    while {[gets $orig line] != -1} {
		if {[regexp {^-+END PGP MESSAGE} $line]} { break }
	    }
	} else {
	    puts $comb $line
	}
   }
    close $orig
    close $comb
    close $mess

    Mh_Rename $outfile $file
    File_Delete $pgpfile

    set msg(dpy) {}
    MsgChange $msg(id)

    return 1
}

# encrypts the current message with the user's key
# added pgp5 support
# works with gpg
proc Pgp_ExmhEncrypt { version } {
    global exmh msg mhProfile pgp pgp5 gpg

    set file $msg(path)
    set tmpfile [Mime_TempFile "encrypt"]

    Exmh_Status "Encrypting $exmh(folder)/$msg(id)"

    Pgp_Process $file $tmpfile "pgp-action: PGP/MIME-encrypt; recipients=[lindex [set ${version}(myname)] 0]; pgp-version=$version"
    Mh_Rename $tmpfile $file

    set msg(dpy) {}
    MsgChange $msg(id)

    return 1
}

# Removes any pgp-action header and puts a brand new one
# added pgp5 support
# works with gpg
proc Pgp_SeditEncrypt { action version draft t } {
    global miscRE pgp pgp5 gpg

    SeditSave $draft $t

    set linenb 1
    set line [$t get $linenb.0 $linenb.end]
    set hasfcc 0

    while {![regexp $miscRE(headerend) $line]} {
	if [regexp -nocase {^pgp-action:} $line] {
	    set line " dummy"
	    while {[regexp "^\[ \t]" $line]} {
		$t delete $linenb.0 [expr {$linenb + 1}].0
		set line [$t get $linenb.0 $linenb.end]
	    }
	} else {
	    if [regexp -nocase {^fcc:} $line] {
		set hasfcc 1
	    }
	    set linenb [expr {$linenb + 1}]
	}
	set line [$t get $linenb.0 $linenb.end]
    }
    if { ![set ${version}(enabled)] } {
	SeditMsg $t "[string toupper $version] Not enabled"
	return
    }
    if {"$action" == "none"} {
        return
    }
    set pgpaction "Pgp-Action: $action"
    if [set ${version}(rfc822)] {
	append pgpaction "; rfc822=on"
    } else {
	append pgpaction "; rfc822=off"
    }
    if [regexp {sign} $action] {
       if { [set ${version}(choosekey)] && \
			[llength [set ${version}(privatekeys)]] > 1} {
	  # choose the private key for the version, you want to use
	  set signkey [Pgp_ChoosePrivateKey $version\
		"Please select the key to use for signing"]
       } else {
	  set signkey [set ${version}(myname)]
       }
       append pgpaction ";\n\toriginator=\"[lindex $signkey 1]\""
    }
    if [regexp {encrypt} $action] {
	catch {
	     append pgpaction ";\n\trecipients=\"[join [Misc_Map key {lindex $key 1} \
		[PgpMatch_Whom $version $draft $hasfcc]] ",\n\t\t    "]\""
	}
    }
    #
    # append PGP version info
    #
    append pgpaction ";\n\tpgp-version=${version}"
    $t insert 1.0 "$pgpaction\n"
}

# added pgp5 support
proc Pgp_ChoosePrivateKey { version text } {
   global pgp pgp5 gpg

   set signkeys {}
   while {[llength $signkeys] != 1} {
      if [catch {Pgp_KeyBox $version $text [set ${version}(secring)] \
		[set ${version}(privatekeys)]} signkeys] {
	 set signkeys [list [set ${version}(myname)]]
      }
   }
   return [lindex $signkeys 0]
}

# added pgp5 support
# works with gpg
proc Pgp_SetMyName { version } {
   global pgp pgp5 gpg

   set ${version}(myname) [Pgp_ChoosePrivateKey $version \
	 "Please select the default key to use for signing"]
}

# added pgp5 support
# works with gpg
proc Pgp_Process { srcfile dstfile {pgpaction {}} } {
    global pgp pgp5 gpg env miscRE

    set orig [open $srcfile r]

    Exmh_Debug Pgp_Process
    # get the header of the draft and split it into mime and non-mime headers
    set mailheaders [Misc_Segregate line \
	  {[regexp $miscRE(mimeheaders) $line]} [Misc_GetHeader $orig]]
    set mimeheaders [lindex $mailheaders 0]
    set mailheaders [lindex $mailheaders 1]
    set hasfcc [expr {[lsearch -glob $mailheaders "fcc:*"] >= 0}]

    # add the default pgp-action header and mime headers
    if {$pgpaction != {}} {
	set mailheaders [linsert $mailheaders 0 $pgpaction]
    }
    if {[lsearch -glob $mimeheaders "content-type:*"] < 0} {
	lappend mimeheaders "content-type: text/plain; charset=us-ascii"
    }

    # get the content transfer encoding (if it's 8bit and we are clear-text
    # signing we should convert to qp before signing).
    if [lsearch -glob $mimeheaders "content-transfer-encoding:*"] {
	set transferencoding [string trim [string tolower [lindex [split \
		[lindex $mimeheaders \
		[lsearch -glob $mimeheaders "content-transfer-encoding:*"]] \
		":"] 1 ]]]
    } else {
	set transferencoding "7bit"
    }

    # get the pgp-action header and parse it
    set pgpaction [split [lindex $mailheaders [lsearch -glob \
					$mailheaders "pgp-action:*"]] ";"]
    set action [string trim [string tolower [lindex \
					[split [lindex $pgpaction 0] ":"] 1]]]
    set actionparams [lrange $pgpaction 1 end]

    # if there is nothing to do, stop here
    if {"$action" == "none"} {
	close $orig
	error "no action"
    }

    # Should we try and apply quoted-printable encoding?
    set doencoding [expr [string match "*signclear" $action] && \
	    [string match "8bit" $transferencoding] && \
	    ![regexp -nocase {^pgp-} $action]]

    # parse pgp-action parameters
    foreach param $actionparams {
	if [regexp {^([^=]+)=(.*)$} $param {} name value] {
	    set actionparam([string tolower [string trim $name]]) \
						[string trim $value " \t\""]
	} else {
	    error "<PGP> incorrect syntax: $param"
	}
    }

    # setup default params
    regsub {.*-} $action {} xaction
    set typeparams "; x-action=$xaction"

    # setup pgp version
    if [info exists actionparam(pgp-version)] {
	set version $actionparam(pgp-version)
    } else {
	set version pgp
    }
    
    # setup rfc822
    if [info exists actionparam(rfc822)] {
	set rfc822 [regexp -nocase $miscRE(true) $actionparam(rfc822)]
    } else {
	set rfc822 [set ${version}(rfc822)]
    }

    # setup the originator (if necessary)
    if [regexp {sign} $action] {
	if [info exists actionparam(originator)] {
	    set originator [PgpMatch_Simple $version \
			$actionparam(originator) [set ${version}(secring)]]
	} else {
	    set originator [set ${version}(myname)]
	}
	append typeparams "; x-originator=[string range [lindex $originator 0] 2 end]"
    }

    # get the ids of the recipients (if necessary)
    if [regexp {encrypt} $action] {
	if [info exists actionparam(recipients)] {
	    set ids [Misc_Map id {PgpMatch_Simple $version $id \
						[set ${version}(pubring)]} \
		    [split $actionparam(recipients) ","]]
	} else {
	    set ids [PgpMatch_Whom $version $srcfile $hasfcc]
	}
	ExmhLog "<Pgp_Process> Encrypting with public key(s): [join $ids ", "]"

	append typeparams ";\n\tx-recipients=\"[join [Misc_Map key \
				{string range [lindex $key 0] 2 end} $ids] ", "]\""
    }
    
    # remove pgp-action and mime-version headers
    set mailheaders [Misc_Filter line \
			{![regexp "^(mime-version|pgp-action):" $line]} $mailheaders]

    # setup the header of the application/pgp subpart
    if $rfc822 {
	set pgpheaders [concat \
		[list "content-type: message/rfc822" ""] \
		[Misc_Filter line {![string match {[bf]cc:*} $line]} 
		$mailheaders] \
		[list "mime-version: 1.0"] \
		$mimeheaders]
    } else {
	set pgpheaders $mimeheaders
    }

    if $doencoding {
	lappend pgpheaders "Content-Transfer-Encoding: quoted-printable"
    } else {
	lappend pgpheaders "Content-Transfer-Encoding: $transferencoding"
    }

    # write the message to be encrypted
    set msgfile [Mime_TempFile "msg"]
    set msg [open $msgfile w 0600]
    
    if ![regexp -nocase {^pgp-} $action] {
	foreach line $pgpheaders { puts $msg [Pgp_FixHeader $line] }
    }
    puts $msg ""

    if $doencoding {
	close $msg
	global mime
	if [catch {open "|$mime(encode) -q >>$msgfile" w} msg] {
	    error "Cannot run $mime(encode)"
	}
    }
    puts -nonewline $msg [read $orig]
    close $orig
    close $msg

    set pgpfile [Mime_TempFile "pgp"]
    
    # each called with version param
    if [catch {
	switch $action {
	    application/pgp-encrypt	{ 
		PgpExec_Encrypt $version $msgfile $pgpfile $ids }
	    application/pgp-signbinary	{
		PgpExec_Sign $version $msgfile $pgpfile $originator 0 }
	    application/pgp-signclear	{ 
		PgpExec_Sign $version $msgfile $pgpfile $originator 1 }
	    application/pgp-encryptsign	{
		PgpExec_EncryptSign $version $msgfile $pgpfile $originator $ids }
	    pgp/mime-signclear { 
		PgpExec_SignPM $version $msgfile $pgpfile $originator }
	    pgp/mime-encryptsign { 
		PgpExec_EncryptSign $version $msgfile $pgpfile $originator $ids}
	    pgp/mime-encrypt	{ PgpExec_Encrypt $version $msgfile $pgpfile $ids }
	    pgp-signclear { 
		PgpExec_Sign $version $msgfile $pgpfile $originator 1}
	    pgp-encryptsign { 
		PgpExec_EncryptSign $version $msgfile $pgpfile $originator $ids}
	    pgp-encrypt	{ PgpExec_Encrypt $version $msgfile $pgpfile $ids }
	    default	{ error "<PGP> unknown action $action" }
	}
    } err] {
	File_Delete $msgfile
	error $err
    }

    # complete mailheaders with the applcation/pgp content-type
    if {[info exists actionparam(localaction)] && \
	    [regexp -nocase $miscRE(true) $actionparam(localaction)]} { 
	set mailheaders {}
    }
	
    switch -regexp $action {
        application/pgp- { 
            lappend mailheaders \
                "mime-version: 1.0" \
                "content-type: application/pgp; format=mime$typeparams" \
                "content-transfer-encoding: 7bit"

            # write out the new mail file
            set dst [open $dstfile w 0600]
            foreach line $mailheaders { puts $dst [Pgp_FixHeader $line] }
            puts $dst ""
            set msg [open $pgpfile r]
            puts -nonewline $dst [read $msg]
            close $msg
            close $dst
        }
        pgp/mime- { 
            Pgp_ProcessPM $version $dstfile $pgpfile $mailheaders $action $msgfile
        }
        pgp- { 
            Pgp_ProcessPlain $dstfile $pgpfile $mailheaders $action $msgfile
        }
    }
    
    
    File_Delete $msgfile $pgpfile
}

# added pgp5 support
# added gpg support
proc Pgp_ProcessPM {version dstfile pgpfile mailheaders action plainfile} {

    # for gpg-digest
    global gpg

    set boundary [Mime_MakeBoundary P]

    # Put in specified headers.
    switch $version {
	pgp  { set micalg md5 }
	pgp5 { set micalg sha1 }
	gpg  { 	#
		# Dont know, if its allowed
		# to specify other micalgs
		# than md5 and sha1 (rfc2015 ?)
		# Just do it!
		set micalg $gpg(digestalgo)
	     }
    }

    lappend mailheaders "mime-version: 1.0"
    if [regexp {encrypt} $action] {
	lappend mailheaders "content-type: multipart/encrypted; boundary=\"$boundary\";\n\t protocol=\"application/pgp-encrypted\""
    } else {
	lappend mailheaders "content-type: multipart/signed; boundary=\"$boundary\";\n\t micalg=pgp-${micalg}; protocol=\"application/pgp-signature\""
    }
    lappend mailheaders "content-transfer-encoding: 7bit"

    # Write file
    set dst [open $dstfile w 0600]
    set pgpIO [open $pgpfile r]

    foreach line $mailheaders { puts $dst [Pgp_FixHeader $line] }
    puts $dst ""
    puts $dst "--$boundary"
    if [regexp {encrypt} $action] {
	puts $dst "Content-Type: application/pgp-encrypted"
	puts $dst ""
	puts $dst "Version: 1"
	puts $dst ""
	puts $dst "--$boundary"
	puts $dst "Content-Type: application/octet-stream"
	puts $dst ""
	puts $dst [read $pgpIO]
	puts $dst "--$boundary--"
    } else {
	set plain [open $plainfile r]
	puts $dst [read $plain]
	close $plain
	puts $dst "--$boundary"
	puts $dst "Content-Type: application/pgp-signature"
	puts $dst ""
	puts $dst [read $pgpIO]
	puts $dst "--$boundary--"
    }

    close $dst
    close $pgpIO

}

# works with pgp5
# works with gpg
proc Pgp_ProcessPlain {dstfile pgpfile mailheaders action plainfile} {

    set dst [open $dstfile w 0600]
    set pgpIO [open $pgpfile r]

    foreach line $mailheaders { puts $dst [Pgp_FixHeader $line] }
    puts $dst ""

    puts $dst [read $pgpIO]

    close $dst
    close $pgpIO
}


##########
#*********
##########


# added pgp5 support
# added gpg support
proc PgpExec_SignPM { version in out sigkey } {

    switch $version {
	pgp  {	PgpExec pgp [list -stab $in -u [lindex $sigkey 0] -o $out]\
		    output $sigkey }
	pgp5 {	PgpExec pgp5 [list pgps -abt $in -u [lindex $sigkey 0] -o $out]\
		    output $sigkey }
	gpg  {  PgpExec gpg [list -abt -u [lindex $sigkey 0] -o $out $in]\
		    output $sigkey }
    }
    if {![file exists $out] && [file exists "$out.asc"]} {
	exec mv "$out.asc" $out
    }
    if {![file exists $out]} {
	error "[string toupper $version] refused to generate the signed text:\n$output"
    } else {
	return {}
    }
}

# Simple version + blank line adding
# added gpg support
proc Pgp_CheckVersion { pgpfile } {
    global pgp pgp5 gpg
    upvar version version
    set in [open $pgpfile r]
    set pgptext [read $in]
    close $in
    #
    # Adding necessary blank lines;
    # Some MS OLE plugins destroy them
    set pgptext [Pgp_CheckBlankLines $pgptext]
    set out [open $pgpfile w]
    puts $out $pgptext
    close $out
    #
    switch -regexp -- $pgptext {
        {Version: 2\.6.*} { if {$pgp(enabled)} { 
				set version pgp
			    } elseif {$pgp5(enabled)} {
				set version pgp5
			    } elseif {$gpg(enabled)} {
				set version gpg
			    }
	}
        {Version: PGP.*(5|6)\..*} { if {$pgp5(enabled)} {
					set version pgp5
				    } elseif {$pgp(enabled)} {
					set version pgp
				    } elseif {$gpg(enabled)} {
					set version gpg
				    }
	}
        {Version: GNUPG} { if {$gpg(enabled)} {
				set version gpg
			   } elseif {$pgp5(enabled)} {
				set version pgp5
			   } elseif {$pgp(enabled)} {
				set version pgp
			   }
	}
    }
}

# Blank line checking and adding, if needed (for some bad OLE messages)
proc Pgp_CheckBlankLines { pgptext } {
    switch -regexp -- $pgptext {
	{BEGIN PGP SIGNED MESSAGE} {
	    #####
	    # Blank line after
	    # -----BEGIN PGP SIGNED MESSAGE-----
	    # Hash: SHA1
	    #####
	    set a [regsub "^(-----BEGIN.+MESSAGE-----\n)(\[^ \n\t:\]+: \[^\n\]+\n)?((\[^ :\t\n]+\[ \t]*)+\n)" $pgptext "\\1\\2\n\\3" pgptext]
	    #####
	    # Blank line after
	    # -----BEGIN PGP SIGNATURE-----
	    # Version: PGP
	    # Charset: noconv
	    # ...    : ...
	    #####
	    set b [regsub "(-----BEGIN.+SIGNATURE-----\n)((\[^ \n\t:\]+: \[^\n\]+\n)+)((\[^ :\t\n]+\[ \t]*)+\n)" $pgptext "\\1\\2\n\\4" pgptext]
	    Exmh_Debug "Pgp_CheckBlankLines: Number of blank lines added: $a-$b"
	    }
	{BEGIN PGP MESSAGE} {
	    #####
	    # Blank line after
	    # -----BEGIN PGP MESSAGE-----
            # Version: PGP
            # Charset: noconv
            # ...    : ...
	    #####
	    set a [regsub "^(-----BEGIN.+MESSAGE-----\n)((\[^ \n\t:\]+: \[^\n\]+\n)+)((\[^ :\t\n]+\[ \t]*)+\n)" $pgptext "\\1\\2\n\\4" pgptext]
	    Exmh_Debug "Pgp_CheckBlankLines: Number of blank lines added: $a"
	    }
    }
    return $pgptext
}


# added pgp5 support
# added gpg support
proc MimeShowMultipartSignedPgp {tkw part} {
    global mimeHdr mime pgp pgp5 gpg

    if [info exists mimeHdr($part=2,file)] {
	# check PGP version
	Pgp_CheckVersion $mimeHdr($part=2,file)
    } else {
	error "Missing signature"
    }

    if { ![info exists version] } {
	error "Signature has no version info"
    }

    if { [set ${version}(enabled)] } {
	
	set PGP [string toupper $version]

	if {![info exists mimeHdr($part,pgpdecode)]} {
	    if {([set ${version}(showinline)] == "all") ||
		([set ${version}(showinline)] == "signed")} {
		set mimeHdr($part,pgpdecode) 1
	    } else { set mimeHdr($part,pgpdecode) 0 }
	}

	MimeMenuAdd $part checkbutton \
	    -label "Check signature with $PGP" \
	    -command [list busy MimeRedisplayPart $tkw $part] \
	    -variable mimeHdr($part,pgpdecode)

	if $mimeHdr($part,pgpdecode) {
	    switch $version {
		pgp  { 	PgpExec $version \
				[list $mimeHdr($part=2,file) \
				[GetSignedText $tkw $part]] msg
		}
		pgp5 {	#
			# pgpv batchmode requires special filenaming
			#
			file copy -force [GetSignedText $tkw $part] \
							[set txtfile /tmp/hugo]
			file copy -force $mimeHdr($part=2,file) \
							[set sigfile /tmp/hugo.asc]
			PgpExec $version [list pgpv $sigfile] msg
			File_Delete $txtfile; File_Delete $sigfile
		}
		gpg {	PgpExec $version \
				[list --verify $mimeHdr($part=2,file) \
				[GetSignedText $tkw $part]] msg
		}
	    }

	    Pgp_InterpretOutput $version $msg pgpresult
	    Exmh_Debug "Key ID: $pgpresult(keyid)"
	    Pgp_DisplayMsg $version $tkw $part pgpresult
    
	    set mimeHdr($part=1,color) \
	        [MimeDarkerColor $tkw $mimeHdr($part,color)]
	} else {
	    $tkw insert insert \
		"$PGP signed message - the signature hasn't been checked\n"
	    TextButton $tkw "Check signature" \
		"$mimeHdr($part,menu) invoke Check\\ signature\\ with\\ $PGP \
                 \n$tkw config -cursor xterm"
	    $tkw insert insert "\n"
	    MimeInsertSeparator $tkw $part 6
	    set mimeHdr($part=1,color) $mimeHdr($part,color)
	}

    } else { set mimeHdr($part=1,color) $mimeHdr($part,color) }

    MimeShowPart $tkw $part=1 [MimeLabel $part part] 1
}

# added pgp5 support
# added gpg support
proc MimeShowMultipartEncryptedPgp {tkw part} {
    global mimeHdr pgp pgp5 gpg exmh

    # check PGP version
    Pgp_CheckVersion $mimeHdr($part=2,file)

    if { ![info exists version] } {
	error "Encrypted Message has no version info"
    }

    if {![set ${version}(enabled)]} {
	Mime_ShowDefault $tkw $part
	return
    }

    set PGP [string toupper $version]

    if {![info exists mimeHdr($part,pgpdecode)]} {
	if {[set ${version}(showinline)] == "all"} {
	    set mimeHdr($part,pgpdecode) 1
	} else { set mimeHdr($part,pgpdecode) 0 }
	
	MimeMenuAdd $part checkbutton \
	    -label "Decrypt with ${PGP}" \
	    -command [list busy MimeRedisplayPart $tkw $part] \
	    -variable mimeHdr($part,pgpdecode)
    }  

    if {!$mimeHdr($part,pgpdecode)} {
	TextButton $tkw "Decrypt $PGP message" \
	    "$mimeHdr($part,menu) invoke Decrypt\\ with\\ ${PGP} \
	     \n$tkw config -cursor xterm"
	$tkw insert insert "\n"
	Mime_ShowDefault $tkw $part
	return
    }
	
    set tmpfile [Mime_TempFile "decrypt"]

    # Decide whether or not to use expect
    if {[set ${version}(keeppass)] && \
		[info exists exmh(expectk)] && [set ${version}(useexpectk)]} {
	# Decrypt version specific
	PgpExec_DecryptExpect $version $mimeHdr($part=2,file) $tmpfile msg
    } else {
	# Assume only recipient is primary secret key
	# Use expect to avoid this behavior
	set recipients [Misc_Map elem {string trim $elem} \
             [split [string range [lindex [set ${version}(myname)] 0] 2 end] ","]]
    	# Decrypt version specific
	PgpExec_Decrypt $version $mimeHdr($part=2,file) $tmpfile msg $recipients
    }

    Pgp_InterpretOutput $version $msg pgpresult
    Pgp_DisplayMsg $version $tkw $part pgpresult

    set DarkerColor [MimeDarkerColor $tkw $mimeHdr($part,color)]
    
    # The following three lines would show the 
    # application/pgp-encrypted mime section.

    # set mimeHdr($part=1,color) $DarkerColor
    # MimeShowPart $tkw $part=1 [MimeLabel $part part] 1
    # MimeInsertSeparator $tkw $part 5
    
    set fileIO [open $tmpfile r]
    set mimeHdr($part=2,color) $DarkerColor
    set mimeHdr($part=2,numParts) [MimeParseSingle $tkw $part=2 $fileIO]
    close $fileIO
    
    MimeShowPart $tkw $part=2=1 [MimeLabel $part part] 1
}

# works with pgp5
# works with gpg
proc GetSignedText {tkw part} {
    global mimeHdr

    set fileIO [open $mimeHdr($part,file) r]
    set boundary $mimeHdr($part,param,boundary)

    # Prolog
    while {([set numBytes [gets $fileIO line]] >= 0) &&
	   ([string compare --$boundary $line] != 0) &&
	   ([string compare --$boundary-- $line] != 0)} {}

    set tmpFilename [Mime_TempFile $part=1]
    set tmpFile [open $tmpFilename w 0600]

    # Body
    # Must not include the newline right before the boundary, that is
    # considered part of the boundary, not the subpart. (rfc1521?)

    # Also, LF should be converted to CRLF for checking.
    
    if {([set numBytes [gets $fileIO line]] >= 0) &&
	([string compare --$boundary $line] != 0) &&
	([string compare --$boundary-- $line] != 0)} {
	puts -nonewline $tmpFile $line
	while {([set numBytes [gets $fileIO line]] >= 0) &&
	       ([string compare --$boundary $line] != 0) &&
	       ([string compare --$boundary-- $line] != 0)} {
	    puts $tmpFile "\r"
	    puts -nonewline $tmpFile $line 
	}
    }
 
    close $fileIO
    close $tmpFile
    
return $tmpFilename
}

# added pgp5 support
# added gpg support
proc Pgp_ShowMessage { tkw part } {
    global mimeHdr mime pgp pgp5 gpg miscRE exmh

    set in [open $mimeHdr($part,file) r]
    gets $in firstLine
    close $in

    # let's get the format
    if {![info exists mimeHdr($part,param,format)]} {
	lappend mimeHdr($part,params) format
	if [regexp $miscRE(beginpgpkeys) $firstLine] {
	    set mimeHdr($part,param,format) keys-only
	} else {
	    set mimeHdr($part,param,format) text
	}
    }
    set format $mimeHdr($part,param,format)

    Exmh_Debug Pgp_ShowMessage format $format $part

    # the action pgp performed
    if {"$format" != "keys-only"} {
	if {![info exists mimeHdr($part,param,x-action)]} {
	    if [regexp $miscRE(beginpgpclear) $firstLine] {
		set action signclear
		set mimeHdr($part,param,x-action) signclear
	    } else {
		set action encryptsign
	    }
	} else {
	    set action $mimeHdr($part,param,x-action)
	}
    } else {
	set action "keys-only"
    }

    # check PGP version
    Pgp_CheckVersion $mimeHdr($part,file)

    # short cut if you don't have PGP at all
    if {(![set ${version}(enabled)]) && ("$action" != "signclear")} {
	Mime_ShowDefault $tkw $part
	return
    }

    catch {set PGP [string toupper $version]}

    # get the recipients if necessary
    if [regexp {encrypt} $action] {
	if {![info exists mimeHdr($part,param,x-recipients)]} {
	    set recipients [string range [lindex [set ${version}(myname)] 0] 2 end]
	} else {
	    set recipients $mimeHdr($part,param,x-recipients)
	}
	set recipients [Misc_Map elem {string trim $elem} [split $recipients ","]]
    }

    # see if we should decode the thing
    if {![info exists mimeHdr($part,pgpdecode)]} {
	set mimeHdr($part,pgpdecode) [expr {[set ${version}(enabled)] && [expr [set ${version}(decode,[set ${version}(showinline)])]]}]
	if [set ${version}(enabled)] {
	    MimeMenuAdd $part checkbutton \
		    -label "[set ${version}(menutext,$action)] with $PGP..." \
		    -command [list busy MimeRedisplayPart $tkw $part] \
		    -variable mimeHdr($part,pgpdecode)
	}
    }
    
    if {($format == "mime") || ($format == "text")} {
	if $mimeHdr($part,pgpdecode) {
	    set tmpfile [Mime_TempFile "decrypt"]
	    
	    if [regexp "encrypt" $action] {
		
		# Decide whether or not to use expect
		if {[set ${version}(keeppass)] && [info exists exmh(expectk)] \
			&& [set ${version}(useexpectk)]} {
		    Exmh_Debug "Using expect"
		    PgpExec_DecryptExpect $version $mimeHdr($part,file) \
			$tmpfile msg
		} else {
		    PgpExec_Decrypt $version $mimeHdr($part,file) \
			$tmpfile msg $recipients
		}
	    } else {
		switch $version {
		    pgp  { PgpExec pgp [list $mimeHdr($part,file) -o $tmpfile] msg }
		    pgp5 { PgpExec pgp5 [list pgpv $mimeHdr($part,file) -o $tmpfile] msg }
		    gpg  { PgpExec gpg [list -o $tmpfile $mimeHdr($part,file)] msg }
		}
	    }

	    Pgp_InterpretOutput $version $msg pgpresult
	    Pgp_DisplayMsg $version $tkw $part pgpresult

	    if {$pgpresult(ok)} {
		if [catch {set fileIO [open $tmpfile r]} err] {
		    Exmh_Debug "Error: $err"
		    return
		}
		File_Delete $tmpfile
    
		if {![info exists mimeHdr($part,numParts)]} {
		    Exmh_Debug MimeParseSingle $part
		    set mimeHdr($part,numParts) \
			[MimeParseSingle $tkw $part $fileIO]
		    set mimeHdr($part=1,color) \
			[MimeDarkerColor $tkw $mimeHdr($part,color)]
		}
		MimeShowPart $tkw $part=1 [MimeLabel $part part] 1
		MimeClose $fileIO
	    }
	    
	} else {
	    if {$action == "signclear"} {
		$tkw insert insert \
		    "$PGP signed message - the signature hasn't been checked\n"
		if [set ${version}(enabled)] {
		    TextButton $tkw "Check signature" \
		    [list $mimeHdr($part,menu) invoke \
			 "[set ${version}(menutext,$action)] with $PGP..."]
		}
		$tkw insert insert "\n"
		MimeInsertSeparator $tkw $part 6
		if [catch {Pgp_Unsign [Misc_FileString $mimeHdr($part,file)]} msg] {
		    $tkw insert insert "  can't find the signed message.\nPlease check it out: it might be suspicious !\n"
		    return
		}
		if {$format == "mime"} {
		    set tmpfile "$mimeHdr($part,file).msg"
		    Misc_StringFile $msg $tmpfile
		    set fileIO [open $tmpfile r]
		    File_Delete $tmpfile
		    if {![info exists mimeHdr($part,numParts)]} {
			set mimeHdr($part,numParts) [MimeParseSingle $tkw $part $fileIO]
			set mimeHdr($part=1,color) [MimeDarkerColor $tkw $mimeHdr($part,color)]
		    }
		    MimeShowPart $tkw $part=1 [MimeLabel $part part] 1
		    MimeClose $fileIO
		} else {
		    $tkw insert insert $msg
		}
	    } else {
		TextButton $tkw "Show with $PGP" \
		    [list $mimeHdr($part,menu) invoke \
			 "[set ${version}(menutext,$action)] with $PGP..."]
		$tkw insert insert "\n"
		Mime_ShowDefault $tkw $part
	    }
	}
    } elseif {$format == "keys-only"} {
	if [set ${version}(autoextract)] {
	    # call with version
	    PgpExec_ExtractKeys $version $mimeHdr($part,file) out 0
	} else {
	    MimeMenuAdd $part command \
		    -label "Extract keys into keyring..." \
		    -command "PgpExec_ExtractKeys $version $mimeHdr($part,file) out"
	    TextButton $tkw "Extract keys" "PgpExec_ExtractKeys $version $mimeHdr($part,file) out"
	    $tkw insert insert "\n"
	}
	if $mimeHdr($part,pgpdecode) {
	    switch $version {
	    	pgp  { 	PgpExec pgp [list $mimeHdr($part,file)] msg }
		pgp5 { 	# There is no pgp5 command to only "view" the contents
			# of a keyfile.
			# In batchmode keys in the keyfile get extracted and
			# added to your keyring automatically
			PgpExec pgp5 [list pgpv $mimeHdr($part,file)] msg }
		gpg  {  PgpExec gpg [list $mimeHdr($part,file)] msg }
	    }
	    #
	    # added pgp5 keytypes
            #
	    regexp "\n(sig|pub|sec|sub|SIG|ret|uid)\[^\n]*" $msg msg
	    $tkw insert insert "$msg\n"
	} else {
	    Mime_ShowDefault $tkw $part
	}
    } else {
	$tkw insert insert "PGP application format '$format' unknown\n"
	return
    }
}

# added pgp5 support
# added gpg support
proc Pgp_InsertKeys { version draft t } {
    global env pgp pgp5 gpg

    if [catch {Pgp_KeyBox $version "select the keys you want to send" \
		[set ${version}(pubring)] \
		[Pgp_FlatKeyList $version "" [set ${version}(secring)]]} keys] {
	SeditMsg $t $keys
	return
    }

    foreach key $keys {
	set keyid [lindex $key 0]
	if {![info exists done($keyid)]} {
	    set done($keyid) 1
	    set tmpfile [Mime_TempFile "insertkeys"]

	    # GPG
	    File_Delete $tmpfile

	    # GPG exits nonstandard (bug?!)
	    #
	    if {($version == "gpg")} {
		catch {PgpExec_GetKeys gpg $keyid $tmpfile} msg
		if {[file exists $tmpfile]} {
		    set filio [open $tmpfile r]
		    if {![regexp {BEGIN PGP PUBLIC KEY BLOCK} [read $filio]]} {
			close $filio
			SeditMsg $t "[string toupper $version] \
                                refuses to generate the key message"
			ExmhLog $msg
			return
		    }
		    close $filio
		} else {
		    SeditMsg $t "[string toupper $version] \
                                refuses to generate the key message"
                    ExmhLog $msg
                    return
		}
	    } else {
		if [catch {PgpExec_GetKeys $version $keyid $tmpfile} msg] {
		    SeditMsg $t "[string toupper $version] \
				refuses to generate the key message"
		    ExmhLog $msg
		    return
		}
	    }

	    # insert key code
	    SeditInsertFile $draft $t $tmpfile 1 7bit {application/pgp; format=keys-only} "keys of [lindex $key 1]"
	    File_Delete $tmpfile
	}
    }
}

# added pgp5 support
# added gpg support
proc Pgp_GetTextAttributes { version summary } {
    global pgp pgp5 gpg

    switch $summary {
	GoodSignatureUntrusted {return [set ${version}(GoodUntrustedSig)]}
	GoodSignatureTrusted   {return [set ${version}(GoodTrustedSig)]}
	BadSignatureTrusted    {return [set ${version}(BadSig)]}
	BadSignatureUntrusted  {return [set ${version}(BadSig)]}
        default                {return [set ${version}(OtherMsg)]}
    }
}

# added pgp5 support
# added gpg support
proc Pgp_DisplayMsg { version tkw part pgpresultvar } {
    global pgp pgp5 gpg
    upvar $pgpresultvar pgpresult

    Exmh_Debug "Pgp_DisplayMsg: $pgpresult(msg)"
    if {[info exists pgpresult(keyid)]} {
	MimeMenuAdd $part command \
	    -label "Query keyserver for key $pgpresult(keyid)" \
	    -command "PgpQueryKey $version $pgpresult(keyid)"
	if {[regexp "PublicMissing" $pgpresult(summary)]} {
	    TextButton $tkw "Query keyserver" \
		"PgpQueryKey $version $pgpresult(keyid)"
	}
	$tkw insert insert "\n"
    }

    set rval [Pgp_GetTextAttributes $version $pgpresult(summary)]
    if {$rval != {}} {
	if [catch {eval {$tkw tag configure PgpResults} $rval} err] {
	    Exmh_Debug tag configure PgpResults $rval: $err
	    $tkw insert insert "$pgpresult(msg)\n"
	} else {
	    $tkw insert insert "$pgpresult(msg)\n" PgpResults
	}
    } else {
	$tkw insert insert "$pgpresult(msg)\n"
    }
    MimeInsertSeparator $tkw $part 6
}
