#
# PGP functions to query for public keys via WWW
# author: Carl Roth <roth@cse.ucsc.edu>
#
# added horowitz key protocol for PGP 5.0
# Markus Gruber <Markus.Gruber@ilo.baynet.de>
#
# query the keyserver:
#

proc PgpQueryKey { version id } {
    global pgp pgp5

    if {$version == "gpg"} {
	Exmh_Status "Sorry, no keyserver support for GPG"
	return
    }
    # dispatch based on whether we use email or WWW:
    switch [set ${version}(keyquerymethod)] {
	hkp { PgpQueryHKPKey $id }
	WWW { PgpQueryWWWKey $version $id }
	email { Misc_Send [set ${version}(keyserver)] "GET 0x$id" }
	default { PgpQueryOtherKey $version $id }
    }
}

# You must have PGP 5.0 to fetch keys via hkp
proc PgpQueryHKPKey { id } {
    global pgp5

    PgpExec pgp5 [list pgpk -a hkp://$pgp5(HKPkeyserverUrl)/0x$id] output {}
    Exmh_Debug $output
    if [regexp {Unable to import keyfile.*} $output msg] {
	Exmh_Status "$msg"
    } else {
	Exmh_Status "Successfully fetched key $id"
    }
}	
    

proc pgp_make_msg {id file} {
    global exmh

    set rcvstore [file dirname $exmh(slocal)]/rcvstore
    set pipe [open "|$rcvstore +$exmh(folder)" w]
    puts $pipe "Mime-Version: 1.0 (generated by exmh)"
    puts $pipe "Content-Type: application/pgp; format=keys-only"
    puts $pipe "Date: [exec date]"
    puts $pipe "From: system"
    puts $pipe "Subject: Output of pgp on id $id"
    puts $pipe ""

    exec cat $file >@$pipe
    close $pipe

    # update the folder listing:
    Flist_FindUnseen
    busy Scan_FolderForce
}

proc PgpQueryStatus {state count length} {
    if {$length} {
	Exmh_Status [format "%s... %.1f%% complete" \
			 $state [expr 100.0 * $count / $length]]
    } else {
	Exmh_Status [format "%s..." $state]
    }
}

proc PgpQueryDone { version url file } {
    global pgp

    upvar #0 $url data
    if {[info exists data(html)]} {
	set fd [open $file w]
	puts -nonewline $fd $data(html)
	close $fd
    }
    set ${version}(querydone) 1
}

# added pgp5 support
proc PgpQueryWWWKey { version key } {
    global pgp pgp5
    set url [format [set ${version}(keyserverURL)] $key]
    set ${version}(querydone) 0
    Exmh_Status "posting query $url"

    set tmp [Mime_TempFile "pgpkey"]
    Http_get $url "PgpQueryDone $version $url $tmp" PgpQueryStatus
    tkwait variable ${version}(querydone)
    if {[file exists $tmp]} {
	Exmh_Status "Saving key $key"
	if [PgpExec_ExtractKeys $version $tmp out 0] {
	    exec rm -f $tmp
	}
    } else {
	Exmh_Status "unable to fetch key!"
    }
}

#
# query a key using an external function
#

# added pgp5 support
proc PgpQueryOtherKey { version id } {
    global env pgp pgp5

    Exmh_Status "querying for key \[$id\]..."

    set tmp [Mime_TempFile "pgpkey"]
    set cmd [eval [set ${version}(keyothermethod)]]
    if {[catch $cmd result]} {
	Exmh_Status [format "unable to get key: %s" \
			 [lindex [split $key "\n"] 0]]
    } else {
	Exmh_Status "filing response"
	pgp_make_msg $id $tmp
    }
    exec rm -f $tmp
}

# added pgp5 support
proc PgpWWW_Init { version } {
    global pgp pgp5

    # for gpg there is no keyserver support yet, so do nothing
    if {$version == "gpg"} {
	return
    }

    set label [string toupper $version]

    switch $version {
	pgp  { set choice [list CHOICE WWW email other] }
	pgp5 { set choice [list CHOICE hkp WWW email other] }
    }

    Preferences_Add "$label interface" {} [list \
	[concat [list ${version}(keyquerymethod) ${version}KeyQueryMethod $choice] [list "Method for querying $label keys"\
"PGP public keys can be queried using email servers
(offline), using WWW servers (interactive) 
or the Horowitz Key Protocol (only with PGP 5.0).
A user-supplied proc (other) can also be given to fetch the key." ] ] ]

    if { $version == "pgp5" } {
	Preferences_Add "$label interface" {} [list \
	    [concat ${version}(HKPkeyserverUrl) ${version}HKPKeyServerUrl {{keys.pgp.com} {Horowitz Key Protocol Server}
"The Horowitz Keyserver directly communicates with the PGP 5.0 pgpk command." }] ]
    }

    Preferences_Add "$label interface" {} [list \
        [concat ${version}(keyserverURL) ${version}KeyServerURL {{http://www-swiss.ai.mit.edu/htbin/pks-extract-key.pl?op=get&search=0x%s} {URL to fetch keys interactively}
"The WWW keyserver is forms based.  Enter a URL here where
the key id (sans 0x) is substituted with %s (using format)" }] \
        [concat ${version}(keyothermethod) ${version}KeyOtherMethod [list {exec echo "can't find $id" > $tmp} "Other method to fetch a $label key"\
"The external method need to fetch the id in \$id and place
it into the temp file \$tmp." ] ] ]

}
